Select the most relevant agent for the user's query.

Each agent can request other agents to fill in the blanks, so pick the agent that can best answer the entire query.

Here's the choice of agents and their uses:
```
{% for agent in agents %}
- `{{ agent.name[:-5] }}`: {{ agent.__doc__.strip().split() | join(' ') }}
{% endfor %}
```
